/*
 * Sonar, open source software quality management tool.
 * Copyright (C) 2009 SonarSource SA
 * mailto:contact AT sonarsource DOT com
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sonar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */

package org.sonar.plugins.cpd;

import net.sourceforge.pmd.cpd.AbstractLanguage;
import net.sourceforge.pmd.cpd.CPD;
import net.sourceforge.pmd.cpd.TokenEntry;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.Predicate;
import org.apache.commons.configuration.Configuration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.sonar.commons.WildcardPattern;
import org.sonar.plugins.api.maven.Exclusions;
import org.sonar.plugins.api.maven.ProjectContext;
import org.sonar.plugins.api.maven.model.MavenPom;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class CpdExecutor {

  private static final Logger LOG = LoggerFactory.getLogger(CpdExecutor.class);

  private MavenPom pom;
  private CpdMapping mapping;
  private Configuration configuration;
  private Exclusions exclusions;
  private ProjectContext context;

  public CpdExecutor(MavenPom pom, ProjectContext context, CpdMapping mapping, Configuration configuration, Exclusions exclusions) {
    this.context = context;
    this.pom = pom;
    this.mapping = mapping;
    this.configuration = configuration;
    this.exclusions = exclusions;
  }

  protected CpdExecutor(MavenPom pom, Exclusions exclusions) {
    this.pom = pom;
    this.exclusions = exclusions;
  }

  public void execute() {
    try {
      LOG.info("Collecting duplications...");

      CPD cpd = newCPD();
      cpd.add(getFilesToInclude(mapping.getLanguage().getFileSuffixes()));
      cpd.go();
      CpdAnalyser cpdAnalyser = new CpdAnalyser(context, mapping, pom.getSourceDirs());

      cpdAnalyser.analyse(cpd.getMatches());
    } catch (Exception e) {
      throw new CpdExecutionException(e);
    }
  }

  private CPD newCPD() {
    // To avoid a cpd bug generating error as "java.lang.IndexOutOfBoundsException: Index: 259, Size: 248"
    // See http://sourceforge.net/tracker/?func=detail&atid=479921&aid=1947823&group_id=56262 for more details
    TokenEntry.clearImages();

    AbstractLanguage cpdLanguage = new AbstractLanguage(mapping.getTokenizer(), mapping.getLanguage().getFileSuffixes().clone()) {
    };
    return new CPD(getMinimumTokenCountConfiguration(), cpdLanguage);
  }

  private Integer getMinimumTokenCountConfiguration() {
    Object minimumTokenCountProperty = configuration.getProperty(CpdPlugin.CPD_MINIMUM_TOKENS_PROPERTY);
    if (minimumTokenCountProperty != null) {
      if (minimumTokenCountProperty instanceof String) {
        return Integer.parseInt((String) minimumTokenCountProperty);
      } else {
        throw new CpdExecutionException("Invalid value of minimum tokens : " + minimumTokenCountProperty);
      }
    } else {
      return Integer.parseInt(CpdPlugin.CPD_MINIMUM_TOKENS_DEFAULT_VALUE);
    }
  }

  protected List<File> getFilesToInclude(String[] suffixes) {
    LOG.info("File suffixes : {}", suffixes);
    if (exclusions != null) {
      LOG.info("Exclusions : {}", exclusions.getWildcardPatterns());
    }

    Collection<File> files = getSourceFilesWithSuffixes(suffixes);
    return new ArrayList<File>(CollectionUtils.select(files, new Predicate() {
      public boolean evaluate(Object object) {
        try {
          File file = (File) object;
          String path = file.getCanonicalPath();
          boolean isAccepted = isPathNotInExclusions(path);
          if (isAccepted) {
            LOG.debug("Including file : {}", path);
          }
          return isAccepted;
        } catch (IOException e) {
          throw new CpdExecutionException(e);
        }
      }
    }));
  }

  private Collection<File> getSourceFilesWithSuffixes(String[] suffixes) {
    return pom.getSourceFiles(suffixes);
  }

  private boolean isPathNotInExclusions(String path) {
    if (exclusions != null) {
      for (String wildcardPattern : exclusions.getWildcardPatterns()) {
        WildcardPattern matcher = new WildcardPattern(wildcardPattern);
        if (matcher.match(path)) {
          return false;
        }
      }
    }
    return true;
  }

}
